@@ -24,6 +24,7 @@
  *      Sanjay Gupta (gsanjay@novell.com)
  *      Mark Steele (ms@rapidsys.com)
  *      Jonathan Gilbert (logic@deltaq.org)
+ *	Sebastien Pouliot  <sebastien@ximian.com>
  *
  * Useful documentation about bitmaps
  *
@@ -75,7 +76,7 @@ gdip_getcodecinfo_bmp ()
         return &bmp_codec; 
 }
 
-GpStatus
+static GpStatus
 gdip_get_bmp_pixelformat (BITMAPINFOHEADER *bih, PixelFormat *dest)
 {
 	int bitCount = bih->biBitCount;
@@ -106,6 +107,10 @@ gdip_get_bmp_pixelformat (BITMAPINFOHEADER *bih, PixelFormat *dest)
 	        case 24:
 	                *dest = Format24bppRgb;
 			break;
+	        case 16:
+			/* MS produce such files (i.e. bad header) for storing ImageList bitmaps, see bug #80797 */
+	                *dest = Format16bppRgb565;
+			break;
 	        case 8:
 	                *dest = Format8bppIndexed;
 			break;
@@ -124,7 +129,7 @@ gdip_get_bmp_pixelformat (BITMAPINFOHEADER *bih, PixelFormat *dest)
 	return Ok;
 }
 
-void 
+static void 
 gdip_bitmap_fill_info_header (GpBitmap *bitmap, PBITMAPINFOHEADER bmi)
 {
 	PixelFormat format = bitmap->active_bitmap->pixel_format;
@@ -160,20 +165,8 @@ gdip_bitmap_fill_info_header (GpBitmap *bitmap, PBITMAPINFOHEADER bmi)
 #endif
 }                                                           
 
-GpStatus 
-gdip_load_bmp_image_from_file (FILE *fp, GpImage **image)
-{
-        return gdip_read_bmp_image_from_file_stream ( (void*)fp, image, TRUE);
-}
-
-GpStatus 
-gdip_load_bmp_image_from_stream_delegate (dstream_t *loader, GpImage **image)
-{
-	return gdip_read_bmp_image_from_file_stream ((void *)loader, image, FALSE);
-}
-
-GpStatus
-gdip_read_bmp_rle_8bit (void *pointer, byte *scan0, bool upsidedown, int stride, int scanWidth, int scanCount, bool useFile)
+static GpStatus
+gdip_read_bmp_rle_8bit (void *pointer, byte *scan0, bool upsidedown, int stride, int scanWidth, int scanCount, ImageSource source)
 {
 	byte code;
 	int bytes_read;
@@ -189,13 +182,13 @@ gdip_read_bmp_rle_8bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 
 	while ((rows_remaining > 0)
 	    || ((row_offset == 0) && (col_offset < scanWidth))) {
-		bytes_read = gdip_read_bmp_data (pointer, &code, 1, useFile);
+		bytes_read = gdip_read_bmp_data (pointer, &code, 1, source);
 
 		if (bytes_read < 1)
 			return GenericError; /* TODO?: Add an "unexpected end of file" error code */
 
 		if (code == 0) { /* RLE escape code */
-			bytes_read = gdip_read_bmp_data (pointer, &code, 1, useFile);
+			bytes_read = gdip_read_bmp_data (pointer, &code, 1, source);
 
 			if (bytes_read < 1)
 				return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -221,8 +214,8 @@ gdip_read_bmp_rle_8bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 				{
 					byte dx, dy;
 
-					bytes_read  = gdip_read_bmp_data (pointer, &dx, 1, useFile);
-					bytes_read += gdip_read_bmp_data (pointer, &dy, 1, useFile);
+					bytes_read  = gdip_read_bmp_data (pointer, &dx, 1, source);
+					bytes_read += gdip_read_bmp_data (pointer, &dy, 1, source);
 
 					if (bytes_read < 2)
 						return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -256,7 +249,7 @@ gdip_read_bmp_rle_8bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 							pointer,
 							&scan0[row_offset + col_offset],
 							bytes_to_read_this_scan,
-							useFile);
+							source);
 
 						if (bytes_read < bytes_to_read_this_scan)
 							return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -280,7 +273,7 @@ gdip_read_bmp_rle_8bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 					}
 
 					if (pad_byte_present) {
-						bytes_read = gdip_read_bmp_data(pointer, &code, 1, useFile);
+						bytes_read = gdip_read_bmp_data(pointer, &code, 1, source);
 
 						if (bytes_read < 1)
 							return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -295,7 +288,7 @@ gdip_read_bmp_rle_8bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 			int run_length = code;
 			byte pixel_value;
 
-			bytes_read = gdip_read_bmp_data(pointer, &pixel_value, 1, useFile);
+			bytes_read = gdip_read_bmp_data(pointer, &pixel_value, 1, source);
 
 			if (bytes_read < 1)
 				return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -331,8 +324,8 @@ gdip_read_bmp_rle_8bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 	return Ok;
 }
 
-GpStatus
-gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride, int scanWidth, int scanCount, bool useFile)
+static GpStatus
+gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride, int scanWidth, int scanCount, ImageSource source)
 {
 	byte code;
 	int bytes_read;
@@ -350,13 +343,13 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 		return InvalidParameter;
 
 	while (rows_remaining > 0) {
-		bytes_read = gdip_read_bmp_data (pointer, &code, 1, useFile);
+		bytes_read = gdip_read_bmp_data (pointer, &code, 1, source);
 
 		if (bytes_read < 1)
 			return GenericError; /* TODO?: Add an "unexpected end of file" error code */
 
 		if (code == 0) { /* RLE escape code */
-			bytes_read = gdip_read_bmp_data (pointer, &code, 1, useFile);
+			bytes_read = gdip_read_bmp_data (pointer, &code, 1, source);
 
 			if (bytes_read < 1)
 				return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -382,8 +375,8 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 				{
 					byte dx, dy;
 
-					bytes_read  = gdip_read_bmp_data (pointer, &dx, 1, useFile);
-					bytes_read += gdip_read_bmp_data (pointer, &dy, 1, useFile);
+					bytes_read  = gdip_read_bmp_data (pointer, &dx, 1, source);
+					bytes_read += gdip_read_bmp_data (pointer, &dy, 1, source);
 
 					if (bytes_read < 2)
 						return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -417,7 +410,7 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 							/* special case: a pair of pixels is split across two rows. */
 							byte pixels, same_row_pixel, next_row_pixel;
 
-							bytes_read = gdip_read_bmp_data (pointer, &pixels, 1, useFile);
+							bytes_read = gdip_read_bmp_data (pointer, &pixels, 1, source);
 
 							if (bytes_read < 1)
 								return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -460,7 +453,7 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 								pointer,
 								&scan0[row_offset + col_offset / 2],
 								bytes_to_read_this_scan,
-								useFile);
+								source);
 
 							if (bytes_read < bytes_to_read_this_scan)
 								return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -482,7 +475,7 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 							while (bytes_to_read_this_scan >= 0) {
 								byte pixels;
 
-								bytes_read = gdip_read_bmp_data (pointer, &pixels, 1, useFile);
+								bytes_read = gdip_read_bmp_data (pointer, &pixels, 1, source);
 
 								if (bytes_read < 1)
 									return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -514,7 +507,7 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 						/* half of a byte remains to be inserted into the correct nybble */
 						byte pixel;
 
-						bytes_read = gdip_read_bmp_data (pointer, &pixel, 1, useFile);
+						bytes_read = gdip_read_bmp_data (pointer, &pixel, 1, source);
 
 						if (bytes_read < 1)
 							return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -545,7 +538,7 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 					}
 
 					if (pad_byte_present) {
-						bytes_read = gdip_read_bmp_data(pointer, &code, 1, useFile);
+						bytes_read = gdip_read_bmp_data(pointer, &code, 1, source);
 
 						if (bytes_read < 1)
 							return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -568,7 +561,7 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 			byte pixel_values;
 			byte inverted_pixel_values;
 
-			bytes_read = gdip_read_bmp_data(pointer, &pixel_values, 1, useFile);
+			bytes_read = gdip_read_bmp_data(pointer, &pixel_values, 1, source);
 
 			if (bytes_read < 1)
 				return GenericError; /* TODO?: Add an "unexpected end of file" error code */
@@ -692,32 +685,32 @@ gdip_read_bmp_rle_4bit (void *pointer, byte *scan0, bool upsidedown, int stride,
 }
 
 GpStatus
-gdip_read_BITMAPINFOHEADER (void *pointer, BITMAPINFOHEADER *bmi, bool useFile, BOOL *os2format, BOOL *upsidedown)
+gdip_read_BITMAPINFOHEADER (void *pointer, BITMAPINFOHEADER *bmi, ImageSource source, BOOL *os2format, BOOL *upsidedown)
 {
 	DWORD dw = 0;
 	byte *data_read = (byte*)&dw;
 	int size = sizeof (DWORD);
-	int size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+	int size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 	if (size_read < size)
 		return InvalidParameter;
 
 	bmi->biSize = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 	if (bmi->biSize > BITMAPCOREHEADER_SIZE){   /* New Windows headers can be bigger */ 
 		dw = 0;
-		size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+		size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 		if (size_read < size)
 			return InvalidParameter;
 		bmi->biWidth = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 
 		dw = 0;
-		size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+		size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 		if (size_read < size)
 			return InvalidParameter;
 		bmi->biHeight = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
  	} else if (bmi->biSize == BITMAPCOREHEADER_SIZE) {
 		/* Old OS/2 format. Width and Height fields are WORDs instead of DWORDS */
 		dw = 0;
-		size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+		size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 		if (size_read < size)
 			return InvalidParameter;
 		bmi->biWidth = (data_read[1]<<8 | data_read[0]);
@@ -728,49 +721,49 @@ gdip_read_BITMAPINFOHEADER (void *pointer, BITMAPINFOHEADER *bmi, bool useFile,
 	}
 
 	dw = 0;
-	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+	size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 	if (size_read < size)
 		return InvalidParameter;
 	bmi->biPlanes = (data_read[1]<<8 | data_read[0]); 
 	bmi->biBitCount = (data_read[3]<<8 | data_read[2]); 
 
 	dw = 0;
-	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+	size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 	if (size_read < size)
 		return InvalidParameter;
 	bmi->biCompression = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
 	if (bmi->biHeight < 0) { /* Negative height indicates that the bitmap is sideup*/
-		upsidedown = FALSE;
+		*upsidedown = FALSE;
 		bmi->biHeight = -bmi->biHeight;
 	}
 
 	dw = 0;
-	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+	size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 	if (size_read < size)
 		return InvalidParameter;
 	bmi->biSizeImage = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
 	dw = 0;
-	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+	size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 	if (size_read < size)
 		return InvalidParameter;
 	bmi->biXPelsPerMeter = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 	
 	dw = 0;
-	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+	size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 	if (size_read < size)
 		return InvalidParameter;
 	bmi->biYPelsPerMeter = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
 	dw = 0;
-	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+	size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 	if (size_read < size)
 		return InvalidParameter;
 	bmi->biClrUsed = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
 		
 	dw = 0;
-	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+	size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 	if (size_read < size)
 		return InvalidParameter;
 	bmi->biClrImportant = (data_read[3]<<24 | data_read[2]<<16 | data_read[1]<<8 | data_read[0]);
@@ -778,10 +771,10 @@ gdip_read_BITMAPINFOHEADER (void *pointer, BITMAPINFOHEADER *bmi, bool useFile,
 	return Ok;
 }
 
+/* For use with in-memory bitmaps, where the BITMAPFILEHEADER doesn't exists */
 GpStatus 
-gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFile)
+gdip_read_bmp_image (void *pointer, GpImage **image, ImageSource source)
 {
-	BITMAPFILEHEADER bmfh;
 	BITMAPINFOHEADER bmi;
 	GpBitmap	*result = NULL;
 	guchar		*pixels = NULL;
@@ -792,7 +785,7 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 	BOOL		upsidedown = TRUE;
 	int		size;
 	int		size_read;
-	byte		*data_read;
+	byte		*data_read = NULL;
 	int		line;
 	int		loop;
 	long		index;
@@ -802,34 +795,8 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 	ARGB green_mask = 0;
 	ARGB blue_mask = 0;
 	int red_shift = 0;
-		
-	size = sizeof(bmfh);
-	data_read = (byte*) GdipAlloc(size);
-	if (data_read == NULL) {
-		status = OutOfMemory;
-		goto error;
-	}
-	memset (data_read, 0, size);
-	size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
-	if (size_read < size) {
-		status = InvalidParameter;
-		goto error;
-	}
-	
-	bmfh.bfType = ((data_read[1]<<8)|data_read[0]);
-	if (bmfh.bfType != BFT_BITMAP) {
-		status = UnknownImageFormat;
-		goto error;
-	}
 
-	bmfh.bfSize = (data_read[5]<<24 | data_read[4]<<16 | data_read[3]<<8 | data_read[2]);
-	bmfh.bfReserved1 = ((data_read[7]<<8)|data_read[6]);
-	bmfh.bfReserved1 = ((data_read[9]<<8)|data_read[8]);
-	bmfh.bfOffBits = (data_read[13]<<24 | data_read[12]<<16 | data_read[11]<<8 | data_read[10]);
-	GdipFree (data_read);
-	data_read = NULL;
-
-	status = gdip_read_BITMAPINFOHEADER (pointer, &bmi, useFile, &os2format, &upsidedown);
+	status = gdip_read_BITMAPINFOHEADER (pointer, &bmi, source, &os2format, &upsidedown);
 	if (status != Ok)
 		goto error;
 
@@ -856,13 +823,13 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 		} else if (bmi.biSize == sizeof (BITMAPINFOHEADER)) {
 			// next three DWORD are the R,G,B masks, like bmiColors in BITMAPINFO
 			int size = sizeof (RGBQUAD);
-			size_read = gdip_read_bmp_data (pointer, (void*)&red_mask, size, useFile);
+			size_read = gdip_read_bmp_data (pointer, (void*)&red_mask, size, source);
 			if (size_read != size)
 				goto error;
-			size_read = gdip_read_bmp_data (pointer, (void*)&green_mask, size, useFile);
+			size_read = gdip_read_bmp_data (pointer, (void*)&green_mask, size, source);
 			if (size_read != size)
 				goto error;
-			size_read = gdip_read_bmp_data (pointer, (void*)&blue_mask, size, useFile);
+			size_read = gdip_read_bmp_data (pointer, (void*)&blue_mask, size, source);
 			if (size_read != size)
 				goto error;
 		}
@@ -874,10 +841,11 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 			/* five red bits, six green bits and five blue bits (0xFFFF) */
 			red_shift = 11;
 		} else {
-			g_warning ("Unsupported 16bbp format with masks A %d, R: %d, G %d, B %d", 
-				alpha_mask, red_mask, green_mask, blue_mask);
-			status = InvalidParameter;
-			goto error;
+			/* MS produce such files (i.e. missing masks) for storing ImageList bitmaps, see bug #80797 */
+			red_mask = 0x7C00;
+			green_mask = 0x3E0;
+			blue_mask = 0x1F;
+			red_shift = 10;
 		}
 
 		/* note: CAIRO_FORMAT_RGB16_565 is deprecated so we're promoting the bitmap to 32RGB */
@@ -945,7 +913,7 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 			goto error;
 		}
 		for (i = 0; i < colours; i++) {
-			size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+			size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 			if (size_read < size) {
 				status = InvalidParameter;
 				goto error;
@@ -970,10 +938,10 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 	if ((bmi.biCompression == BI_RLE4) || (bmi.biCompression == BI_RLE8)) {
 		switch (bmi.biCompression) {
 			case BI_RLE4:
-				gdip_read_bmp_rle_4bit (pointer, pixels, upsidedown, result->active_bitmap->stride, result->active_bitmap->width, result->active_bitmap->height, useFile);
+				gdip_read_bmp_rle_4bit (pointer, pixels, upsidedown, result->active_bitmap->stride, result->active_bitmap->width, result->active_bitmap->height, source);
 				break;
 			case BI_RLE8:
-				gdip_read_bmp_rle_8bit (pointer, pixels, upsidedown, result->active_bitmap->stride, result->active_bitmap->width, result->active_bitmap->height, useFile);
+				gdip_read_bmp_rle_8bit (pointer, pixels, upsidedown, result->active_bitmap->stride, result->active_bitmap->width, result->active_bitmap->height, source);
 				break;
 		}
 	} else {
@@ -1019,11 +987,18 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 				line = i;
 			}
 
-			size_read = gdip_read_bmp_data (pointer, data_read, size, useFile);
+			size_read = gdip_read_bmp_data (pointer, data_read, size, source);
 
 			if (size_read < size) {
-				status = InvalidParameter;
-				goto error;
+				/* special case for missing data in the last line */
+				if (line == result->active_bitmap->height - 1) {
+					/* MS produce such files for storing ImageList bitmaps, see bug #80797 */
+					int missing_size = size - size_read;
+					memset (data_read + size_read, 0, missing_size);
+				} else {
+					status = InvalidParameter;
+					goto error;
+				}
 			}
 
 			switch(bmi.biBitCount) {
@@ -1050,38 +1025,20 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 					continue;
 				}
 
-				case 24: {
+				case 24:
+				case 32: {
 					int src = 0;
 					int dest = 0;
+					int skip = (bmi.biBitCount >> 3);
 
 					index = (line * result->active_bitmap->stride);
 					while (src < loop) {
 						set_pixel_bgra(pixels, index + dest, data_read[src+0], data_read[src+1], data_read[src+2], 0xff);
 						dest += 4;
-						src += 3;
+						src += skip;
 					}
 					continue;
 				}
-
-				case 32: {
-#ifndef WORDS_BIGENDIAN
-					memcpy (pixels + line * result->active_bitmap->stride, data_read, loop);
-#else
-					int	src;
-					int	dest;
-
-					src = 0;
-					dest = 0;
-
-					while (src < loop) {
-						index = (line * result->active_bitmap->stride);
-						set_pixel_bgra(pixels, index+dest, data_read[src+0], data_read[src+1], data_read[src+2], data_read[src+3]);
-						dest += 4;
-						src += 4;
-					}
-#endif
-					continue;
-				}
 			}
 		}
 
@@ -1110,15 +1067,68 @@ error:
 	}
 
 	return InvalidParameter;
+}
 
+/* BMP read from files have a BITMAPFILEHEADER but this isn't the case for the GDI API
+ * (e.g. displaying a bitmap) */
+static GpStatus 
+gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, ImageSource source)
+{
+	GpStatus status;
+	BITMAPFILEHEADER bmfh;
+	int size_read;
+		
+	int size = sizeof (bmfh);
+	BYTE* data_read = (BYTE*) GdipAlloc (size);
+
+	if (!data_read)
+		return OutOfMemory;
+
+	memset (data_read, 0, size);
+	size_read = gdip_read_bmp_data (pointer, data_read, size, source);
+	if (size_read < size) {
+		status = InvalidParameter;
+		goto error;
+	}
+	
+	bmfh.bfType = ((data_read[1]<<8)|data_read[0]);
+	if (bmfh.bfType != BFT_BITMAP) {
+		status = UnknownImageFormat;
+		goto error;
+	}
+
+	bmfh.bfSize = (data_read[5]<<24 | data_read[4]<<16 | data_read[3]<<8 | data_read[2]);
+	bmfh.bfReserved1 = ((data_read[7]<<8)|data_read[6]);
+	bmfh.bfReserved1 = ((data_read[9]<<8)|data_read[8]);
+	bmfh.bfOffBits = (data_read[13]<<24 | data_read[12]<<16 | data_read[11]<<8 | data_read[10]);
+	GdipFree (data_read);
+
+	return gdip_read_bmp_image (pointer, image, source);
+error:
+	if (data_read != NULL)
+		GdipFree(data_read);
+	return status;
+}
+
+GpStatus 
+gdip_load_bmp_image_from_file (FILE *fp, GpImage **image)
+{
+        return gdip_read_bmp_image_from_file_stream ((void*)fp, image, File);
+}
+
+GpStatus 
+gdip_load_bmp_image_from_stream_delegate (dstream_t *loader, GpImage **image)
+{
+	return gdip_read_bmp_image_from_file_stream ((void*)loader, image, DStream);
 }
 
 int 
-gdip_read_bmp_data (void *pointer, byte *data, int size, bool useFile)
+gdip_read_bmp_data (void *pointer, byte *data, int size, ImageSource source)
 {
-	if (useFile) {
+	switch (source) {
+	case File:
 		return fread (data, 1, size, (FILE*) pointer);
-	} else {
+	case DStream: {
 		/* Streams are not required to return the number of bytes
 		   requested, they could return less yet our code seems to assume
 		   it will always get what it's asking for; lets loop until we
@@ -1140,22 +1150,30 @@ gdip_read_bmp_data (void *pointer, byte *data, int size, bool useFile)
 
 		return total;
 	}
+	case Memory: {
+		MemorySource *ms = (MemorySource*)pointer;
+		int len = (ms->pos + size < ms->size) ? size : ms->size - ms->pos;
+		if (len > 0) {
+			memcpy (data, ms->ptr + ms->pos, len);
+			ms->pos += len;
+		}
+		return size;
+	}
+	default:
+		return -1;
+	}
 }
 
-GpStatus 
-gdip_save_bmp_image_to_file (FILE *fp, GpImage *image)
+static void 
+gdip_write_bmp_data (void *pointer, byte *data, int size, bool useFile)
 {
-	return gdip_save_bmp_image_to_file_stream ( (void *)fp, image, TRUE);
+	if (useFile)
+		fwrite (data, 1, size, (FILE*) pointer);
+	else 	
+		((PutBytesDelegate)(pointer))(data, size);	
 }
 
-GpStatus 
-gdip_save_bmp_image_to_stream_delegate (PutBytesDelegate putBytesFunc,
-                                        GpImage *image)
-{	
-        return gdip_save_bmp_image_to_file_stream ( (void *)putBytesFunc, image, FALSE);
-}
-
-GpStatus 
+static GpStatus 
 gdip_save_bmp_image_to_file_stream (void *pointer, GpImage *image, bool useFile)
 {
 	BITMAPFILEHEADER	bmfh;
@@ -1291,13 +1309,15 @@ gdip_save_bmp_image_to_file_stream (void *pointer, GpImage *image, bool useFile)
 	return Ok;
 }
 
-void 
-gdip_write_bmp_data (void *pointer, byte *data, int size, bool useFile)
+GpStatus 
+gdip_save_bmp_image_to_file (FILE *fp, GpImage *image)
 {
-	if (useFile)
-		fwrite (data, 1, size, (FILE*) pointer);
-	else 	
-		((PutBytesDelegate)(pointer))(data, size);	
+	return gdip_save_bmp_image_to_file_stream ( (void *)fp, image, TRUE);
 }
 
-
+GpStatus 
+gdip_save_bmp_image_to_stream_delegate (PutBytesDelegate putBytesFunc,
+                                        GpImage *image)
+{	
+        return gdip_save_bmp_image_to_file_stream ( (void *)putBytesFunc, image, FALSE);
+}
