@@ -10,8 +10,9 @@
  *	Vladimir Vukicevic (vladimir@pobox.com)
  *	Geoff Norton (gnorton@customerdna.com)
  *      Jonathan Gilbert (logic@deltaq.org)
+ *	Sebastien Pouliot  <sebastien@ximian.com>
  *
- * Copyright (C) 2003-2006 Novell, Inc (http://www.novell.com)
+ * Copyright (C) 2003-2007 Novell, Inc (http://www.novell.com)
  */
 
 #ifndef _GDIP_H
@@ -26,37 +27,61 @@
 
 #include "pixman.h"
 
-#ifdef USE_INCLUDED_CAIRO
-#include <cairo.h>
+#if HAVE_VISIBILITY_HIDDEN
+	#define GDIP_INTERNAL __attribute__ ((visibility ("hidden")))
 #else
-#include <cairo/cairo.h>
+	#define GDIP_INTERNAL 
 #endif
-#include <X11/Xlib.h>
 
-#ifdef CAIRO_HAS_FT_FONT
 #ifdef USE_INCLUDED_CAIRO
-#include <cairo-ft.h>
+	#include "cairo-embed.h"
+	#include "cairo.h"
+
+	#ifdef CAIRO_HAS_FT_FONT
+		#include "cairo-ft.h"
+	#endif
+
+	#ifdef CAIRO_HAS_XLIB_SURFACE
+		#include "cairo-xlib.h"
+	#endif
 #else
-#include <cairo/cairo-ft.h>
-#endif
+	#include <cairo/cairo.h>
+	#ifdef CAIRO_HAS_FT_FONT
+		#include <cairo/cairo-ft.h>
+	#endif
+	#ifdef CAIRO_HAS_XLIB_SURFACE
+		#include <cairo/cairo-xlib.h>
+	#endif
 #endif
 
+#include <X11/Xlib.h>
+
 #define gdip_cairo_matrix_copy(m1, m2) memcpy (m1, m2, sizeof (cairo_matrix_t))
 
 #ifdef WORDS_BIGENDIAN
 #define set_pixel_bgra(pixel,index,b,g,r,a) do { \
-		((unsigned char *)(pixel))[index+0] = a; \
+		((unsigned char *)(pixel))[index]   = a; \
 		((unsigned char *)(pixel))[index+1] = r; \
 		((unsigned char *)(pixel))[index+2] = g; \
 		((unsigned char *)(pixel))[index+3] = b; \
 	} while (0)
+#define set_pixel_bgr(pixel,index,b,g,r) do { \
+		((unsigned char *)(pixel))[index]   = r; \
+		((unsigned char *)(pixel))[index+1] = g; \
+		((unsigned char *)(pixel))[index+2] = b; \
+	} while (0)
 #else
 #define set_pixel_bgra(pixel,index,b,g,r,a) do { \
-		((unsigned char *)(pixel))[index+0] = b; \
+		((unsigned char *)(pixel))[index]   = b; \
 		((unsigned char *)(pixel))[index+1] = g; \
 		((unsigned char *)(pixel))[index+2] = r; \
 		((unsigned char *)(pixel))[index+3] = a; \
 	} while (0)
+#define set_pixel_bgr(pixel,index,b,g,r) do { \
+		((unsigned char *)(pixel))[index]   = b; \
+		((unsigned char *)(pixel))[index+1] = g; \
+		((unsigned char *)(pixel))[index+2] = r; \
+	} while (0)
 #endif
 #define get_pixel_bgra(color, b, g, r, a) do { \
 		a = ((color & 0xff000000) >> 24); \
@@ -65,13 +90,6 @@
 		b = (color & 0x000000ff); \
 	} while(0)
 
-#ifdef CAIRO_HAS_XLIB_SURFACE
-#ifdef USE_INCLUDED_CAIRO
-#include <cairo-xlib.h>
-#else
-#include <cairo/cairo-xlib.h>
-#endif
-#endif
 
 /* mono/io-layer/uglify.h also has these typedefs.
  * To avoid a dependency on mono we have copied all
@@ -118,9 +136,6 @@ typedef gpointer HRGN;
 #define DOUBLE_TO_16_16(d) ((FT_Fixed)((d) * 65536.0))
 #define DOUBLE_FROM_16_16(t) ((double)(t) / 65536.0)
 
-struct cairo_matrix {
-	double m[3][2];
-};
 /*
  * Callbacks
  */
@@ -143,6 +158,11 @@ typedef ImageAbort GetThumbnailImageAbort;
 #define CAIRO_AA_OFFSET_X	1
 #define CAIRO_AA_OFFSET_Y	0.5
 
+#define COLOR_PALETTE_FLAGS_DEFAULT	0
+#define COLOR_PALETTE_FLAGS_HAS_ALPHA	1
+#define COLOR_PALETTE_FLAGS_GREYSCALE	2
+#define COLOR_PALETTE_FLAGS_HALFTONE	4
+
 typedef unsigned char byte;
 typedef int bool;
 typedef guint16 WCHAR; /* 16-bits unicode */
@@ -600,12 +620,6 @@ typedef enum {
         RegionTypePath  = 3
 } RegionType;
 
-typedef enum {
-	X11DRV_GET_DISPLAY,			/* get X11 display for a DC */
-	X11DRV_GET_DRAWABLE,		/* get current drawable for a DC */
-	X11DRV_GET_FONT			/* get current X font for a DC */
-} x11drv_escape_codes;
-
 
 /*
  * Structures
@@ -615,6 +629,7 @@ typedef enum {
 #define GBD_OWN_SCAN0	(1<<8)
 #define GBD_WRITE_OK	(1<<9)
 #define GBD_LOCKED	(1<<10)
+#define GBD_TRUE24BPP	(1<<11)
 
 typedef struct {
 	int X, Y, Width, Height;
@@ -908,15 +923,6 @@ typedef struct {
     char lfFaceName[LF_FACESIZE];
 } LOGFONTA;
  
-typedef struct
-{
-	char		szFamily [128];
-	float		sizeInPixels;
-	GpFontStyle	style;
-	GpFont *	font;
-	int		refcount;
-} GpCachedFont;
-
 typedef struct {
         StringAlignment alignment;
         StringAlignment lineAlignment;
@@ -931,11 +937,6 @@ typedef struct {
 	int charRangeCount;
 } GpStringFormat;
 
-typedef struct {
-        float X, Y;
-        float width;
-} GpLinePointF;
-
 typedef struct {
 	GpColorMap *colormap;
 	int colormap_elem;
@@ -995,60 +996,46 @@ typedef struct
 	unsigned char	*scan;
 } StreamingState;
 
-GpStatus gdip_init_pixel_stream (StreamingState *state, BitmapData *data, int x, int y, int w, int h);
-BOOL gdip_pixel_stream_has_next (StreamingState *state);
-unsigned int gdip_pixel_stream_get_next (StreamingState *state);
-void gdip_pixel_stream_set_next (StreamingState *state, unsigned int pixel_value);
+GpStatus gdip_init_pixel_stream (StreamingState *state, BitmapData *data, int x, int y, int w, int h) GDIP_INTERNAL;
+unsigned int gdip_pixel_stream_get_next (StreamingState *state) GDIP_INTERNAL;
 
 /*
  * Functions
  * 
  */
-void gdip_image_init              (GpImage *image);
+void gdip_image_init (GpImage *image) GDIP_INTERNAL;
+void gdip_bitmap_init (GpBitmap *bitmap) GDIP_INTERNAL;
 
-void		gdip_bitmap_init(GpBitmap *bitmap);
-GpBitmap	*gdip_bitmap_new(void);
-GpBitmap	*gdip_bitmap_new_with_frame(const GUID *dimension, bool add_bitmapdata);
-FrameData	*gdip_frame_add(GpBitmap *bitmap, const GUID *dimension);
-BitmapData	*gdip_frame_add_bitmapdata(FrameData *frame);
-void		gdip_bitmap_dispose (GpBitmap *bitmap);
-GpStatus	gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap);
-GpStatus	gdip_bitmap_setactive(GpBitmap *bitmap, const GUID *dimension, int index);
-GpStatus	gdip_bitmapdata_clone(BitmapData *src, BitmapData **dest, int count);
-ColorPalette	*gdip_palette_clone(ColorPalette *original);
-GpStatus	gdip_property_get_short(int offset, void *value, unsigned short *result);
-GpStatus	gdip_property_get_long(int offset, void *value, guint32 *result);
-GpStatus	gdip_property_get_srational(int offset, void *value, unsigned short *numerator, unsigned short *denominator);
-GpStatus	gdip_property_get_rational(int offset, void *value, guint32 *numerator, guint32 *denominator);
-GpStatus	gdip_bitmapdata_property_add(BitmapData *bitmap_data, PROPID id, ULONG length, WORD type, VOID *value);
-GpStatus	gdip_bitmapdata_property_add_byte(BitmapData *bitmap_data, PROPID id, byte value);
-GpStatus	gdip_bitmapdata_property_add_short(BitmapData *bitmap_data, PROPID id, unsigned short value);
-GpStatus	gdip_bitmapdata_property_add_long(BitmapData *bitmap_data, PROPID id, guint32 value);
-GpStatus	gdip_bitmapdata_property_add_ASCII(BitmapData *bitmap_data, PROPID id, unsigned char *value);
-GpStatus	gdip_bitmapdata_property_add_rational(BitmapData *bitmap_data, PROPID id, guint32 numerator, guint32 denominator);
-GpStatus	gdip_bitmapdata_property_add_srational(BitmapData *bitmap_data, PROPID id, unsigned short numerator, unsigned short denominator);
-GpStatus	gdip_bitmapdata_property_remove_id(BitmapData *bitmap_data, PROPID id);
-GpStatus	gdip_bitmapdata_property_remove_index(BitmapData *bitmap_data, int index);
-GpStatus	gdip_bitmapdata_property_find_id(BitmapData *bitmap_data, PROPID id, int *index);
+GpBitmap	*gdip_bitmap_new (void) GDIP_INTERNAL;
+GpBitmap	*gdip_bitmap_new_with_frame (const GUID *dimension, bool add_bitmapdata) GDIP_INTERNAL;
+FrameData	*gdip_frame_add(GpBitmap *bitmap, const GUID *dimension) GDIP_INTERNAL;
+BitmapData	*gdip_frame_add_bitmapdata(FrameData *frame) GDIP_INTERNAL;
+GpStatus	gdip_bitmap_dispose (GpBitmap *bitmap) GDIP_INTERNAL;
+GpStatus	gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap) GDIP_INTERNAL;
+GpStatus	gdip_bitmap_setactive (GpBitmap *bitmap, const GUID *dimension, int index) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_clone (BitmapData *src, BitmapData **dest, int count) GDIP_INTERNAL;
+ColorPalette	*gdip_palette_clone(ColorPalette *original) GDIP_INTERNAL;
+GpStatus	gdip_property_get_short (int offset, void *value, unsigned short *result) GDIP_INTERNAL;
+GpStatus	gdip_property_get_long (int offset, void *value, guint32 *result) GDIP_INTERNAL;
+GpStatus	gdip_property_get_srational (int offset, void *value, unsigned short *numerator, unsigned short *denominator) GDIP_INTERNAL;
+GpStatus	gdip_property_get_rational (int offset, void *value, guint32 *numerator, guint32 *denominator) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_add (BitmapData *bitmap_data, PROPID id, ULONG length, WORD type, VOID *value) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_add_byte (BitmapData *bitmap_data, PROPID id, byte value) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_add_short (BitmapData *bitmap_data, PROPID id, unsigned short value) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_add_long (BitmapData *bitmap_data, PROPID id, guint32 value) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_add_ASCII (BitmapData *bitmap_data, PROPID id, unsigned char *value) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_add_rational (BitmapData *bitmap_data, PROPID id, guint32 numerator, guint32 denominator) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_add_srational (BitmapData *bitmap_data, PROPID id, unsigned short numerator, unsigned short denominator) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_remove_id (BitmapData *bitmap_data, PROPID id) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_remove_index (BitmapData *bitmap_data, int index) GDIP_INTERNAL;
+GpStatus	gdip_bitmapdata_property_find_id (BitmapData *bitmap_data, PROPID id, int *index) GDIP_INTERNAL;
 
-
-void gdip_graphics_init (GpGraphics *graphics, cairo_surface_t *surface);
-GpGraphics *gdip_graphics_new (cairo_surface_t *surface);
-
-GpStatus gdip_bitmap_clone_data_rect (GdipBitmapData *srcData, Rect *srcRect, GdipBitmapData *destData, Rect *destRect);
-GpStatus gdip_bitmap_change_rect_pixel_format (GdipBitmapData *srcData, Rect *srcRect, GdipBitmapData *destData, Rect *destRect);
-
-int gdip_from_RGB_to_ARGB (BYTE *src, int width, int height, int stride, BYTE **dest, int* dest_stride);
-int gdip_from_ARGB_to_RGB (BYTE *src, int width, int height, int stride, BYTE **dest, int* dest_stride);
+GpGraphics *gdip_graphics_new (cairo_surface_t *surface) GDIP_INTERNAL;
 
 /* Pen */
-void gdip_pen_init (GpPen *pen);
-GpPen *gdip_pen_new (void);
-GpStatus gdip_pen_setup (GpGraphics *graphics, GpPen *pen);
+GpStatus gdip_pen_setup (GpGraphics *graphics, GpPen *pen) GDIP_INTERNAL;
 
-GpStatus initCodecList (void);
-void releaseCodecList (void);
-void gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes* attr, bool *allocated);
+void gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes* attr, bool *allocated) GDIP_INTERNAL;
 
 
 /* Bitmap */
@@ -1060,7 +1047,8 @@ GpStatus GdipBitmapLockBits (GpBitmap *bmp, Rect *rc, int flags, int format, Gdi
 GpStatus GdipBitmapUnlockBits (GpBitmap *bmp, GdipBitmapData *bmpData);
 GpStatus GdipBitmapSetPixel ( GpBitmap *bmp, int x, int y, ARGB color);
 GpStatus GdipBitmapGetPixel ( GpBitmap *bmp, int x, int y, ARGB *color);
-
+GpStatus GdipCloneBitmapAreaI (int x, int y, int width, int height, PixelFormat format, GpBitmap *original, GpBitmap **bitmap);
+GpStatus GdipCloneBitmapArea (float x, float y, float w, float h, PixelFormat format, GpBitmap *original, GpBitmap **bitmap);
 
 /* Graphics */
 GpStatus GdipBeginContainer (GpGraphics *graphics, GDIPCONST GpRectF* dstrect, GDIPCONST GpRectF *srcrect, GpUnit unit, GpGraphicsContainer *state);
@@ -1072,8 +1060,15 @@ GpStatus GdipGetDC (GpGraphics *graphics, void **hDC);
 GpStatus GdipReleaseDC (GpGraphics *graphics, void *hDC);
 GpStatus GdipRestoreGraphics (GpGraphics *graphics, unsigned int graphicsState);
 GpStatus GdipSaveGraphics(GpGraphics *graphics, unsigned int * state);
+GpStatus GdipGetWorldTransform (GpGraphics *graphics, GpMatrix *matrix);
+GpStatus GdipMultiplyWorldTransform (GpGraphics *graphics, GpMatrix *matrix, GpMatrixOrder order);
+GpStatus GdipResetWorldTransform (GpGraphics *graphics);
+GpStatus GdipSetWorldTransform (GpGraphics *graphics, GpMatrix *matrix);
+GpStatus GdipScaleWorldTransform (GpGraphics *graphics, float sx, float sy, GpMatrixOrder order);
 GpStatus GdipRotateWorldTransform (GpGraphics *graphics, float angle, GpMatrixOrder order);
 GpStatus GdipTranslateWorldTransform (GpGraphics *graphics, float dx, float dy, GpMatrixOrder order);
+GpStatus GdipDrawArc (GpGraphics *graphics, GpPen *pen, float x, float y, float width, float height, float startAngle, float sweepAngle);
+GpStatus GdipDrawArcI (GpGraphics *graphics, GpPen *pen, int x, int y, int width, int height, float startAngle, float sweepAngle);
 GpStatus GdipDrawBezier (GpGraphics *graphics, GpPen *pen, float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4);
 GpStatus GdipDrawBezierI (GpGraphics *graphics, GpPen *pen, int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4);
 GpStatus GdipDrawBeziers (GpGraphics *graphics, GpPen *pen, GpPointF *points, int count);
@@ -1105,6 +1100,7 @@ GpStatus GdipDrawLine (GpGraphics *graphics, GpPen *pen, float x1, float y1, flo
 GpStatus GdipDrawLineI (GpGraphics *graphics, GpPen *pen, int x1, int y1, int x2, int y2);
 GpStatus GdipDrawLines (GpGraphics *graphics, GpPen *pen, GpPointF *points, int count);
 GpStatus GdipDrawLinesI (GpGraphics *graphics, GpPen *pen, GpPoint *points, int count);
+GpStatus GdipDrawPath (GpGraphics *graphics, GpPen *pen, GpPath *path);
 GpStatus GdipDrawPie (GpGraphics *graphics, GpPen *pen, float x, float y, float width, float height, float startAngle, float sweepAngle);
 GpStatus GdipDrawPieI (GpGraphics *graphics, GpPen *pen, int x, int y, int width, int height, float startAngle, float sweepAngle);
 GpStatus GdipDrawPolygon (GpGraphics *graphics, GpPen *pen, GpPointF *points, int count);
@@ -1158,7 +1154,7 @@ GpStatus GdipTranslateClip (GpGraphics *graphics, float dx, float dy);
 GpStatus GdipTranslateClipI (GpGraphics *graphics, int dx, int dy);
 
 /* Status */
-GpStatus gdip_get_status (cairo_status_t status);
+GpStatus gdip_get_status (cairo_status_t status) GDIP_INTERNAL;
 
 /* Pen */
 GpStatus GdipCreatePen1 (int argb, float width, GpUnit unit, GpPen **pen);
@@ -1282,7 +1278,7 @@ GpStatus GdipPrivateAddFontFile(GpFontCollection* fontCollection,  GDIPCONST WCH
 
 
 /* Font */
-GpStatus GdipCreateFont(GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style, Unit unit,  GpFont **font);
+GpStatus GdipCreateFont(GDIPCONST GpFontFamily* family, float emSize, int style, Unit unit,  GpFont **font);
 GpStatus GdipCreateFontFamilyFromName(GDIPCONST WCHAR *name, GpFontCollection *fontCollection, GpFontFamily **FontFamily);
 GpStatus GdipDeleteFont(GpFont* font);
 
@@ -1374,10 +1370,9 @@ GpStatus GdipGetImageEncoders (UINT numEncoders, UINT size, ImageCodecInfo *enco
 #include "graphics-path.h"
 
 /* for drawing curves */
-GpPointF *convert_points (const GpPoint *points, int count);
-GpPointF *gdip_closed_curve_tangents (int terms, const GpPointF *points, int count, float tension);
-GpPointF *gdip_open_curve_tangents (int terms,  const GpPointF *points, int count, float tension);
-void gdip_calculate_coefficients (int count, int terms, float **coefficients, int *coefficients_count);
+GpPointF *convert_points (const GpPoint *points, int count) GDIP_INTERNAL;
+GpPointF *gdip_closed_curve_tangents (int terms, const GpPointF *points, int count, float tension) GDIP_INTERNAL;
+GpPointF *gdip_open_curve_tangents (int terms,  const GpPointF *points, int count, float tension) GDIP_INTERNAL;
 
 /* Memory */
 void *GdipAlloc (int size);
@@ -1386,30 +1381,20 @@ void *GdipCalloc (size_t nelem, size_t elsize);
 void GdipFree (void *ptr);
 
 /* Utility*/
-int fcmp (double x1, double x2, double epsilon);
-float gdip_get_display_dpi();
-float gdip_unitx_convgr (GpGraphics *graphics, float nSrc);
-float gdip_unity_convgr (GpGraphics *graphics, float nSrc);
-float gdip_unit_conversion (Unit from, Unit to, float dpi, GraphicsType type, float nSrc);
+float gdip_get_display_dpi () GDIP_INTERNAL;
+float gdip_unitx_convgr (GpGraphics *graphics, float nSrc) GDIP_INTERNAL;
+float gdip_unity_convgr (GpGraphics *graphics, float nSrc) GDIP_INTERNAL;
+float gdip_unit_conversion (Unit from, Unit to, float dpi, GraphicsType type, float nSrc) GDIP_INTERNAL;
 
-cairo_content_t from_cairoformat_to_content (cairo_format_t format);
+cairo_surface_t * gdip_bitmap_ensure_surface (GpBitmap *bitmap) GDIP_INTERNAL;
+BOOL gdip_is_an_indexed_pixelformat (PixelFormat pixfmt) GDIP_INTERNAL;
+GpBitmap * gdip_convert_indexed_to_rgb (GpBitmap *bitmap) GDIP_INTERNAL;
 
-void gdip_rect_expand_by (GpRectF *rect, GpPointF *point);
+const EncoderParameter *gdip_find_encoder_parameter (GDIPCONST EncoderParameters *eps, const GUID *guid) GDIP_INTERNAL;
 
-cairo_surface_t * gdip_bitmap_ensure_surface (GpBitmap *bitmap);
-BOOL gdip_is_an_alpha_pixelformat (PixelFormat pixfmt);
-BOOL gdip_is_an_indexed_pixelformat (PixelFormat pixfmt);
-GpBitmap * gdip_convert_indexed_to_rgb (GpBitmap *bitmap);
+GpStatus gdip_get_pattern_status (cairo_pattern_t *pat) GDIP_INTERNAL;
 
-const EncoderParameter *gdip_find_encoder_parameter (GDIPCONST EncoderParameters *eps, const GUID *guid);
-gchar *ucs2_to_utf8(const gunichar2 *ucs2, int length);
-bool utf8_to_ucs2(const gchar *utf8, gunichar2 *ucs2, int ucs2_len);
-int utf8_encode_ucs2char(gunichar2 unichar, unsigned char *dest);
-int utf8_decode_ucs2char(const unsigned char *src, gunichar2 *uchar);
-
-GpStatus gdip_get_pattern_status (cairo_pattern_t *pat);
-
-void gdip_font_clear_pattern_cache (void);
+void gdip_font_clear_pattern_cache (void) GDIP_INTERNAL;
 
 
 /* Stream handling bits */
