@@ -1,7 +1,7 @@
 /*
  * pathgradientbrush.c
  *
- * Copyright (C) Novell, Inc. 2003-2004.
+ * Copyright (C) 2003-2004,2007 Novell, Inc. http://www.novell.com
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
  * and associated documentation files (the "Software"), to deal in the Software without restriction, 
@@ -24,7 +24,7 @@
  *
  */
 
-#include "gdip.h"
+#include "general.h"
 #include "pathgradientbrush.h"
 
 static GpStatus gdip_pgrad_setup (GpGraphics *graphics, GpBrush *brush);
@@ -36,7 +36,7 @@ static BrushClass pathgradient_vtable = { BrushTypePathGradient,
                                           gdip_pgrad_clone_brush,
                                           gdip_pgrad_destroy };
 
-void
+static void
 gdip_pathgradient_init (GpPathGradient *pg)
 {
 	gdip_brush_init (&pg->base, &pathgradient_vtable);
@@ -67,7 +67,7 @@ gdip_pathgradient_init (GpPathGradient *pg)
 	pg->pattern = NULL;
 }
 
-GpPathGradient*
+static GpPathGradient*
 gdip_pathgradient_new (void)
 {
 	GpPathGradient *result = (GpPathGradient *) GdipAlloc (sizeof (GpPathGradient));
@@ -185,12 +185,10 @@ SUCCESS:
 GpStatus
 gdip_pgrad_destroy (GpBrush *brush)
 {
-	GpPathGradient *pgbrush;
+	/* a. the NULL check for brush is done by the caller, GdipDeleteBrush */
+	/* b. brush itself is freed by the caller */
 
-	if (!brush)
-		return InvalidParameter;
-
-	pgbrush = (GpPathGradient *) brush;
+	GpPathGradient *pgbrush = (GpPathGradient *) brush;
 
 	if (pgbrush->boundary) {
 		GdipDeletePath (pgbrush->boundary);
@@ -225,7 +223,6 @@ gdip_pgrad_destroy (GpBrush *brush)
 		pgbrush->presetColors = NULL;
 	}
 
-	GdipFree (pgbrush);
 	return Ok;
 }
 
@@ -353,7 +350,7 @@ gdip_pgrad_setup (GpGraphics *graphics, GpBrush *brush)
 	return gdip_get_status (cairo_status (graphics->ct));
 }
 
-GpPointF
+static GpPointF
 gdip_get_center (GDIPCONST GpPointF *points, int count)
 {
 	/* Center is the mean of all the points. */
@@ -385,6 +382,33 @@ gdip_has_non_empty_color (GDIPCONST ARGB *colors, int count)
 	return FALSE;
 }
 
+static void
+gdip_rect_expand_by (GpRectF *rect, GpPointF *point)
+{
+	/* This method is somewhat stupid, because GpRect is x,y width,height,
+	* instead of x0,y0 x1,y1.
+	*/
+	float x0 = rect->X;
+	float y0 = rect->Y;
+	float x1 = x0 + rect->Width;
+	float y1 = y0 + rect->Height;
+
+	if (point->X < x0)
+		x0 = point->X;
+	else if (point->X > x1)
+		x1 = point->X;
+
+	if (point->Y < y0)
+		y0 = point->Y;
+	else if (point->Y > y1)
+		y1 = point->Y;
+
+	rect->X = x0;
+	rect->Y = y0;
+	rect->Width = (x1 - x0);
+	rect->Height = (y1 - y0);
+}
+
 /* coverity[+alloc : arg-*3] */
 GpStatus
 GdipCreatePathGradient (GDIPCONST GpPointF *points, int count, GpWrapMode wrapMode, GpPathGradient **polyGradient)
@@ -761,9 +785,6 @@ GdipSetPathGradientPresetBlend (GpPathGradient *brush, GDIPCONST ARGB *blend, GD
 	return Ok;
 }
 
-/* This is defined in general.c */
-float gdip_erf (float x, float std, float mean);
-
 GpStatus
 GdipSetPathGradientSigmaBlend (GpPathGradient *brush, float focus, float scale)
 {
