@@ -29,7 +29,7 @@
 #include "gdip.h"
 #include "solidbrush.h"
 
-void 
+static void 
 gdip_pen_init (GpPen *pen)
 {
         pen->color = 0;
@@ -54,7 +54,7 @@ gdip_pen_init (GpPen *pen)
 	cairo_matrix_init_identity (&pen->matrix);
 }
 
-GpPen*
+static GpPen*
 gdip_pen_new (void)
 {
         GpPen *result = (GpPen *) GdipAlloc (sizeof (GpPen));
@@ -466,12 +466,10 @@ GdipSetPenBrushFill (GpPen *pen, GpBrush *brush)
 GpStatus
 GdipGetPenBrushFill (GpPen *pen, GpBrush **brush)
 {
-	if (!pen || !brush)
+	if (!pen)
 		return InvalidParameter;
 
-        *brush = pen->brush;
-
-        return Ok;
+	return GdipCloneBrush (pen->brush, brush);
 }
 
 GpStatus
@@ -494,9 +492,13 @@ GdipSetPenColor (GpPen *pen, int argb)
 	if (!pen)
 		return InvalidParameter;
 
-	pen->changed = (pen->color != argb);
-        pen->color = argb;
-        return Ok;
+	pen->changed = pen->changed ? TRUE : (pen->color != argb);
+	pen->color = argb;
+
+	if (pen->changed && pen->brush && (pen->brush->vtable->type == BrushTypeSolidColor))
+		return GdipSetSolidFillColor ((GpSolidFill*)pen->brush, argb);
+
+	return Ok;
 }
 
 GpStatus
@@ -518,7 +520,7 @@ GdipSetPenMiterLimit (GpPen *pen, float miterLimit)
 	if (miterLimit < 1.0f)
 		miterLimit = 1.0f;
 
-	pen->changed = (pen->miter_limit != miterLimit);
+	pen->changed = pen->changed ? TRUE : (pen->miter_limit != miterLimit);
         pen->miter_limit = miterLimit;
         return Ok;
 }
@@ -539,7 +541,7 @@ GdipSetPenLineJoin (GpPen *pen, GpLineJoin lineJoin)
 	if (!pen)
 		return InvalidParameter;
 
-	pen->changed = (pen->line_join != lineJoin);
+	pen->changed = pen->changed ? TRUE : (pen->line_join != lineJoin);
         pen->line_join = lineJoin;
         return Ok;
 }
@@ -584,7 +586,7 @@ GdipSetPenMode (GpPen *pen, GpPenAlignment penMode)
 	if (!pen)
 		return InvalidParameter;
 
-	pen->changed = (penMode != pen->mode);
+	pen->changed = pen->changed ? TRUE : (penMode != pen->mode);
 	pen->mode = penMode;
 	return Ok;
 }
