@@ -75,7 +75,7 @@ gdip_getcodecinfo_png ()
 #define siglongjmp longjmp
 #endif
 
-void
+static void
 _gdip_png_stream_read_data (png_structp png_ptr, png_bytep data, png_size_t length)
 {
 	GetBytesDelegate	getBytesFunc;
@@ -95,23 +95,22 @@ _gdip_png_stream_read_data (png_structp png_ptr, png_bytep data, png_size_t leng
 	}
 }
 
-void
+static void
 _gdip_png_stream_write_data (png_structp png_ptr, png_bytep data, png_size_t length)
 {
     PutBytesDelegate putBytesFunc = (PutBytesDelegate) png_get_io_ptr (png_ptr);
     putBytesFunc (data, length);
 }
 
-void
+static void
 _gdip_png_stream_flush_data (png_structp png_ptr)
 {
     /* nothing */
 }
 
-GpStatus
-gdip_load_png_properties(png_structp png_ptr, png_infop info_ptr, png_infop end_ptr, BitmapData *bitmap_data)
+static GpStatus
+gdip_load_png_properties (png_structp png_ptr, png_infop info_ptr, png_infop end_ptr, BitmapData *bitmap_data)
 {
-
 #if defined(PNG_INCH_CONVERSIONS) && defined(PNG_FLOATING_POINT_SUPPORTED)
 	bitmap_data->image_flags |= ImageFlagsHasRealDPI;
 	bitmap_data->dpi_horz = png_get_x_pixels_per_inch(png_ptr, info_ptr);
@@ -221,7 +220,7 @@ gdip_load_png_properties(png_structp png_ptr, png_infop info_ptr, png_infop end_
 	return Ok;
 }
 
-GpStatus 
+static GpStatus 
 gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc, GpImage **image)
 {
 	png_structp	png_ptr = NULL;
@@ -229,6 +228,7 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 	png_infop	end_info_ptr = NULL;
 	guchar		*rawdata = NULL;
 	GpImage		*result = NULL;
+	GpStatus	status = InvalidParameter;
 
 	png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
 
@@ -407,9 +407,17 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 
 		width = png_get_image_width (png_ptr, info_ptr);
 		height = png_get_image_height (png_ptr, info_ptr);
-		bit_depth = png_get_bit_depth (png_ptr, info_ptr);
-		color_type = png_get_color_type (png_ptr, info_ptr);
 		channels = png_get_channels (png_ptr, info_ptr);
+
+		/* 24 and 32bpp are supported, 48, 64bpp aren't, see http://bugzilla.ximian.com/show_bug.cgi?id=80693 */
+		bit_depth = png_get_bit_depth (png_ptr, info_ptr);
+		if (bit_depth > 8) {
+			g_warning ("PNG images with %dbpp aren't supported by libgdiplus.", channels * bit_depth);
+			status = UnknownImageFormat;
+			goto error;
+		}
+
+		color_type = png_get_color_type (png_ptr, info_ptr);
 		interlace = png_get_interlace_type (png_ptr, info_ptr);
 
 		/* According to the libpng manual, this sequence is equivalent to
@@ -538,7 +546,7 @@ error:
 	}
 
 	*image = NULL;
-	return InvalidParameter;
+	return status;
 }
 
 
@@ -554,7 +562,7 @@ gdip_load_png_image_from_stream_delegate (GetBytesDelegate getBytesFunc, SeekDel
 	return gdip_load_png_image_from_file_or_stream (NULL, getBytesFunc, image);
 }
 
-GpStatus 
+static GpStatus 
 gdip_save_png_image_to_file_or_stream (FILE *fp, PutBytesDelegate putBytesFunc, GpImage *image, GDIPCONST EncoderParameters *params)
 {
 	png_structp	png_ptr = NULL;
