@@ -1,7 +1,7 @@
 /*
  * texturebrush.c
  *
- * Copyright (C) 2003,2006 Novell, Inc. http://www.novell.com
+ * Copyright (C) 2003,2006-2007 Novell, Inc. http://www.novell.com
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
  * and associated documentation files (the "Software"), to deal in the Software without restriction, 
@@ -42,7 +42,21 @@ static BrushClass vtable = { BrushTypeTextureFill,
 			     gdip_texture_clone,
 			     gdip_texture_destroy };
 
-void 
+static cairo_content_t
+from_cairoformat_to_content (cairo_format_t format)
+{
+    	switch (format) {
+    	case CAIRO_FORMAT_RGB24:
+		return CAIRO_CONTENT_COLOR;
+	case CAIRO_FORMAT_A8:
+		return CAIRO_CONTENT_ALPHA;
+	case CAIRO_FORMAT_ARGB32:
+	default:
+		return CAIRO_CONTENT_COLOR_ALPHA;
+    }
+}
+
+static void 
 gdip_texture_init (GpTexture *texture)
 {
 	gdip_brush_init (&texture->base, &vtable);
@@ -55,7 +69,7 @@ gdip_texture_init (GpTexture *texture)
 	cairo_matrix_init_identity (&texture->matrix);
 }
 
-GpTexture *
+static GpTexture*
 gdip_texture_new (void)
 {
         GpTexture *result = (GpTexture *) GdipAlloc (sizeof (GpTexture));
@@ -70,7 +84,7 @@ gdip_texture_new (void)
 /* 
  * functions to create different wrapmodes.
  */
-GpStatus
+static GpStatus
 draw_tile_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 {
 	cairo_surface_t *original;
@@ -126,7 +140,7 @@ draw_tile_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	return gdip_get_status (cairo_status (ct));
 }
 
-GpStatus
+static GpStatus
 draw_tile_flipX_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 {
 	cairo_surface_t *original;
@@ -196,7 +210,7 @@ draw_tile_flipX_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	return gdip_get_status (cairo_status (ct));
 }
 
-GpStatus
+static GpStatus
 draw_tile_flipY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 {
 	cairo_surface_t *original;
@@ -266,7 +280,7 @@ draw_tile_flipY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	return gdip_get_status (cairo_status (ct));
 }
 
-GpStatus
+static GpStatus
 draw_tile_flipXY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 {
 	cairo_surface_t *original;
@@ -363,7 +377,7 @@ draw_tile_flipXY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	return gdip_get_status (cairo_status (ct));
 }
 
-GpStatus
+static GpStatus
 draw_clamp_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 {
 	cairo_surface_t *original;
@@ -568,13 +582,10 @@ gdip_texture_clone (GpBrush *brush, GpBrush **clonedBrush)
 GpStatus
 gdip_texture_destroy (GpBrush *brush)
 {
-	GpTexture *texture;
+	/* a. the NULL check for brush is done by the caller, GdipDeleteBrush */
+	/* b. brush itself is freed by the caller */
 
-	if (brush == NULL) {
-		return InvalidParameter;
-	}
-
-	texture = (GpTexture *) brush;
+	GpTexture *texture = (GpTexture *) brush;
 
 	if (texture->pattern) {
 		cairo_pattern_destroy (texture->pattern);
@@ -586,8 +597,6 @@ gdip_texture_destroy (GpBrush *brush)
 		texture->image = NULL;
 	}
 
-	GdipFree (texture);
-
 	return Ok;
 }
 
@@ -595,7 +604,7 @@ gdip_texture_destroy (GpBrush *brush)
 GpStatus
 GdipCreateTexture (GpImage *image, GpWrapMode wrapMode, GpTexture **texture)
 {
-	cairo_surface_t *imageSurface;
+	cairo_surface_t *imageSurface = NULL;
 	GpTexture	*result;
 	GpStatus status;
 
@@ -609,34 +618,24 @@ GdipCreateTexture (GpImage *image, GpWrapMode wrapMode, GpTexture **texture)
 		return NotImplemented;
 	}
 
-	imageSurface = cairo_image_surface_create_for_data ((unsigned char *)image->active_bitmap->scan0,
-					image->cairo_format, image->active_bitmap->width,
-					image->active_bitmap->height, image->active_bitmap->stride);
-	if (imageSurface == NULL) {
-		return OutOfMemory;
-	}
-
 	result = gdip_texture_new ();
-	if (result == NULL) {
-		cairo_surface_destroy (imageSurface);
+	if (!result)
 		return OutOfMemory;
-	}
 
 	result->image = NULL;
 	status = GdipCloneImage (image, &result->image);
-	if (status != Ok) {
-		if (result->image)
-			GdipDisposeImage (result->image);
-		cairo_surface_destroy (imageSurface);
-		GdipFree (result);
-		*texture = NULL;
-		return status;
-	}
+	if (status != Ok)
+		goto failure;
+
+	/* note: we must keep the scan0 alive, so we must use the cloned image (and not the original) see bug #80971 */
+	imageSurface = cairo_image_surface_create_for_data ((unsigned char *)result->image->active_bitmap->scan0,
+		image->cairo_format, image->active_bitmap->width, image->active_bitmap->height, image->active_bitmap->stride);
+	if (!imageSurface)
+		goto failure;
 
 	result->wrapMode = wrapMode;
-	if (result->image->surface != NULL) {
-		cairo_surface_destroy(result->image->surface);
-	}
+	if (result->image->surface)
+		cairo_surface_destroy (result->image->surface);
 	result->image->surface = imageSurface;
 	result->rectangle.X = 0;
 	result->rectangle.Y = 0;
@@ -645,6 +644,15 @@ GdipCreateTexture (GpImage *image, GpWrapMode wrapMode, GpTexture **texture)
 
 	*texture = result;
 	return Ok;
+
+failure:
+	if (result->image)
+		GdipDisposeImage (result->image);
+	if (imageSurface)
+		cairo_surface_destroy (imageSurface);
+	GdipFree (result);
+	*texture = NULL;
+	return status;
 }
 
 /* coverity[+alloc : arg-*6] */
@@ -658,20 +666,14 @@ GdipCreateTexture2 (GpImage *image, GpWrapMode wrapMode, float x, float y, float
 GpStatus
 GdipCreateTexture2I (GpImage *image, GpWrapMode wrapMode, int x, int y, int width, int height, GpTexture **texture)
 {
-	int		bmpWidth;
-	int		bmpHeight;
-	cairo_t		*ct;
-	cairo_surface_t *original;
-	cairo_surface_t	*new;
-	GpTexture	*result;
+	int bmpWidth;
+	int bmpHeight;
 	GpStatus status;
+	GpImage *resized_image = NULL;
 
 	if (!image || !texture)
 		return InvalidParameter;
 
-	if ((wrapMode < WrapModeTile) || (wrapMode > WrapModeClamp))
-		return OutOfMemory;
-
 	if (image->type != imageBitmap)
 		return NotImplemented;
 
@@ -682,78 +684,13 @@ GdipCreateTexture2I (GpImage *image, GpWrapMode wrapMode, int x, int y, int widt
 	if ((x < 0) || (y < 0) || (width <= 0) || (height <= 0) || (bmpWidth < (x + width)) || (bmpHeight < (y + height)))
 		return OutOfMemory;
 
-	original = cairo_image_surface_create_for_data ((unsigned char *)image->active_bitmap->scan0, 
-			image->cairo_format, x + width, y + height, image->active_bitmap->stride);
-	if (!original)
-		return OutOfMemory;
-
-	/* texture surface to be used by brush */
-	new = cairo_surface_create_similar (original, from_cairoformat_to_content (image->cairo_format), width, height);
-
-	if (new == NULL) {
-		cairo_surface_destroy (original);
-		return OutOfMemory;
-	}
-
-	/* clip the rectangle from original image surface and create new surface */
-	ct = cairo_create (new);
-	cairo_translate (ct, -x, -y);
-	cairo_set_source_surface (ct, original, x + width, y + height);
-	cairo_paint (ct);
-	cairo_destroy (ct);
-	cairo_surface_destroy (original);
-
-	result = gdip_texture_new ();
-	if (!result) {
-		cairo_surface_destroy (new);
-		return OutOfMemory;
-	}
-
-	result->image = NULL;
-	if ((x == 0) && (y == 0)) {
-		/* the texture share the same dimension as the image, so we can clone it*/
-		status = GdipCloneImage (image, &result->image);
-	} else {
-		/* otherwise we need to make a copy of the requested (rectangle) region */
-		int new_stride = (int) (image->active_bitmap->stride * (image->active_bitmap->width - x) / image->active_bitmap->width);
-		int format = -1;
-		switch (image->cairo_format) {
-		case CAIRO_FORMAT_RGB24:
-			format = Format24bppRgb;
-			break;
-		case CAIRO_FORMAT_ARGB32:
-			format = Format32bppArgb;
-			break;
-		case CAIRO_FORMAT_A8:
-			format = Format8bppIndexed;
-			break;
-		case CAIRO_FORMAT_A1:
-			format = Format1bppIndexed;
-			break;
-		}
-		status = GdipCreateBitmapFromScan0 (width, height, new_stride, format, 
-			(unsigned char *)image->active_bitmap->scan0, &result->image);
-	}
-	if (status != Ok) {
-		if (result->image)
-			GdipDisposeImage (result->image);
-		cairo_surface_destroy (new);
-		GdipFree (result);
-		*texture = NULL;
+	status = GdipCloneBitmapAreaI (x, y, width, height, image->active_bitmap->pixel_format, image, &resized_image);
+	if (status != Ok)
 		return status;
-	}
 
-	if (result->image->surface)
-		cairo_surface_destroy (result->image->surface);
-	result->image->surface = new;
-	result->wrapMode = wrapMode;
-	result->rectangle.X = x;
-	result->rectangle.Y = y;
-	result->rectangle.Width = width;
-	result->rectangle.Height = height;
-
-	*texture = result;
-	return Ok;
+	status = GdipCreateTexture (resized_image, wrapMode, texture);
+	GdipDisposeImage (resized_image);
+	return status;
 }
 
 /* coverity[+alloc : arg-*6] */
