@@ -59,6 +59,8 @@ static const WCHAR tiff_codecname[] = {'B', 'u', 'i','l', 't', '-','i', 'n', ' '
 static const WCHAR tiff_extension[] = {'*', '.', 'T', 'I', 'F',';', '*', '.', 'T', 'I', 'F','F', 0}; /* *.TIF;*.TIFF */
 static const WCHAR tiff_mimetype[] = {'i', 'm', 'a','g', 'e', '/', 't', 'i', 'f', 'f', 0}; /* image/gif */
 static const WCHAR tiff_format[] = {'T', 'I', 'F', 'F', 0}; /* TIFF */
+static const BYTE tiff_sig_pattern[] = { 0x49, 0x49, 0x4D, 0x4D };
+static const BYTE tiff_sig_mask[] = { 0xFF, 0xFF, 0xFF, 0xFF };
 
 /*Wrapper functions and client data strucutre for delegate functions from StreamHelper class of 
 System.Drawing Namespace*/
@@ -70,25 +72,25 @@ typedef struct {
         SizeDelegate sizeFunc;
 } gdip_tiff_clientData;
 
-tsize_t 
+static tsize_t 
 gdip_tiff_fileread (thandle_t clientData, tdata_t buffer, tsize_t size)
 {
 	return (tsize_t)fread(buffer, 1, size, (FILE*)clientData);
 }
 
-tsize_t 
+static tsize_t 
 gdip_tiff_fileread_none (thandle_t clientData, tdata_t buffer, tsize_t size)
 {
 	return 0;
 }
 
-tsize_t 
+static tsize_t 
 gdip_tiff_filewrite (thandle_t clientData, tdata_t buffer, tsize_t size)
 {
 	return (tsize_t)fwrite (buffer, 1, size, (FILE*)clientData);
 }
 
-toff_t 
+static toff_t 
 gdip_tiff_fileseek (thandle_t clientData, toff_t offSet, int whence)
 {
 	int seek_ok = fseek ((FILE*)clientData, offSet, whence);
@@ -98,59 +100,60 @@ gdip_tiff_fileseek (thandle_t clientData, toff_t offSet, int whence)
 	return -1;
 }
 
-int 
+static int 
 gdip_tiff_fileclose (thandle_t clientData)
 {
 	/* This is a TIFF cleanup function; but we own the FILE* and close it in image.c so this is a null op */
 	return 0;
 }
 
-toff_t 
+static toff_t 
 gdip_tiff_filesize (thandle_t clientData)
 {
+	long ret;
 	long cur_pos = ftell ((FILE*)clientData);
 	fseek ((FILE*)clientData, 0, SEEK_END);
-	long ret = ftell ((FILE*)clientData);
+	ret = ftell ((FILE*)clientData);
 	fseek ((FILE*)clientData, cur_pos, SEEK_SET);
 	return (toff_t)ret;
 }
 
-int
+static int
 gdip_tiff_filedummy_map (thandle_t clientData, tdata_t *phase, toff_t* size)
 {
 	return 0;
 }
 
-void
+static void
 gdip_tiff_filedummy_unmap (thandle_t clientData, tdata_t base, toff_t size)
 {
 }
 
-tsize_t 
+static tsize_t 
 gdip_tiff_read (thandle_t clientData, tdata_t buffer, tsize_t size)
 {
 	return (tsize_t)((gdip_tiff_clientData *) clientData)->getBytesFunc (buffer, size, 0);
 }
 
-tsize_t 
+static tsize_t 
 gdip_tiff_read_none (thandle_t clientData, tdata_t buffer, tsize_t size)
 {
 	return 0;
 }
 
-tsize_t 
+static tsize_t 
 gdip_tiff_write (thandle_t clientData, tdata_t buffer, tsize_t size)
 {
 	return (tsize_t)((gdip_tiff_clientData *) clientData)->putBytesFunc (buffer, size);
 }
 
-toff_t 
+static toff_t 
 gdip_tiff_seek (thandle_t clientData, toff_t offSet, int whence)
 {
 	return (toff_t)((gdip_tiff_clientData *) clientData)->seekFunc (offSet, whence);
 }
 
-int 
+static int 
 gdip_tiff_close (thandle_t clientData)
 {
 	/* We should not close the user provided streams */
@@ -158,19 +161,19 @@ gdip_tiff_close (thandle_t clientData)
 	return 1;
 }
 
-toff_t 
+static toff_t 
 gdip_tiff_size (thandle_t clientData)
 {
 	return (toff_t)((gdip_tiff_clientData *) clientData)->sizeFunc ();
 }
 
-int
+static int
 gdip_tiff_dummy_map (thandle_t clientData, tdata_t *phase, toff_t* size)
 {
 	return 0;
 }
 
-void
+static void
 gdip_tiff_dummy_unmap (thandle_t clientData, tdata_t base, toff_t size)
 {
 }
@@ -187,10 +190,10 @@ gdip_getcodecinfo_tiff ()
 	tiff_codec.MimeType = (const WCHAR*) tiff_mimetype;
 	tiff_codec.Flags = Encoder | Decoder | SupportBitmap | Builtin;
 	tiff_codec.Version = 1;
-	tiff_codec.SigCount = 0;
-	tiff_codec.SigSize = 0;
-	tiff_codec.SigPattern = 0;
-	tiff_codec.SigMask = 0;
+	tiff_codec.SigCount = 2;
+	tiff_codec.SigSize = 2;
+	tiff_codec.SigPattern = tiff_sig_pattern;
+	tiff_codec.SigMask = tiff_sig_mask;
 
 	/* Mute the lib */
 	TIFFSetErrorHandler(NULL);
@@ -199,8 +202,8 @@ gdip_getcodecinfo_tiff ()
 	return &tiff_codec;
 }
 
-GpStatus
-gdip_load_tiff_properties(TIFF *tiff, BitmapData *bitmap_data)
+static GpStatus
+gdip_load_tiff_properties (TIFF *tiff, BitmapData *bitmap_data)
 {
 	unsigned char	*text;
 	uint32	i;
@@ -212,7 +215,7 @@ gdip_load_tiff_properties(TIFF *tiff, BitmapData *bitmap_data)
 	uint16	samples_per_pixel;
 	uint16	bits_per_sample;
 	uint16	planar_configuration;
-	uint16	image_length;
+	uint32	image_length;
 	uint16	strips_per_image;
 	uint32	rows_per_strip;
 	uint32	tile_length;
@@ -653,8 +656,8 @@ gdip_load_tiff_properties(TIFF *tiff, BitmapData *bitmap_data)
 	return Ok;
 }
 
-GpStatus
-gdip_save_tiff_properties(TIFF *tiff, BitmapData *bitmap_data, int samples_per_pixel, int bits_per_sample)
+static GpStatus
+gdip_save_tiff_properties (TIFF *tiff, BitmapData *bitmap_data, int samples_per_pixel, int bits_per_sample)
 {
 	int		index;
 	unsigned char	*text;
@@ -696,7 +699,7 @@ gdip_save_tiff_properties(TIFF *tiff, BitmapData *bitmap_data, int samples_per_p
 	if (gdip_bitmapdata_property_find_id(bitmap_data, ExtraSamples, &index) == Ok) {
 		TIFFSetField(tiff, TIFFTAG_EXTRASAMPLES, bitmap_data->property[index].length, bitmap_data->property[index].value);
 	} else {
-		TIFFSetField (tiff, TIFFTAG_EXTRASAMPLES, EXTRASAMPLE_UNSPECIFIED);
+		TIFFSetField (tiff, TIFFTAG_EXTRASAMPLES, 1, EXTRASAMPLE_UNSPECIFIED);
 	}
 
 	if (gdip_bitmapdata_property_find_id(bitmap_data, FillOrder, &index) == Ok) {
@@ -940,7 +943,7 @@ gdip_save_tiff_properties(TIFF *tiff, BitmapData *bitmap_data, int samples_per_p
 }
 
 /*TODO Handle TIFF Encoder Parameters*/
-GpStatus 
+static GpStatus 
 gdip_save_tiff_image (TIFF* tiff, GpImage *image, GDIPCONST EncoderParameters *params)
 {
 	int		frame;
@@ -1053,7 +1056,7 @@ error:
 }
 
 
-GpStatus 
+static GpStatus 
 gdip_load_tiff_image (TIFF *tiff, GpImage **image)
 {
 	int		i;
